package robot;

import ilog.concert.IloException;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumExpr;
import ilog.concert.IloNumVar;
import ilog.concert.IloObjective;
import ilog.cplex.IloCplex;
import ilog.cplex.IloCplexMultiCriterionExpr;
import static robot.Objective.DELAY;
import static robot.Objective.DISTANCE;
import static robot.Objective.DISTANCE_DELAY;

/**
 * Model implements an integer programming model for the routing problem.
 *
 * @author Paul A. Rubin (rubin@msu.edu)
 */
public final class Model implements AutoCloseable {
  private static final double HALF = 0.5;   // rounding tolerance for binaries
  private final Problem prob;               // the underlying problem
  private final int nNodes;                 // number of nodes
  private final int maxMoves;               // maximum number of robot moves
  private final IloCplex model;             // the IP model
  private final IloNumVar[][] location;     // location[t, i] = 1 if the robot
                                            // is at node i at time t
  private final IloNumVar[][] first;        // first[t, i] = 1 if the robot
                                            // first visits node i at time t
  private final IloNumVar[] done;           // done[t] = 1 if the robot returns
                                            // to node 0 for good at time t
  private final IloLinearNumExpr distance;  // distance objective
  private final IloLinearNumExpr delay;     // delay objective
  private final Objective objective;        // the choice of objective(s)
  private IloObjective modelObjective;      // the objective function in us

  /**
   * Constructs the model.
   * @param problem the problem to solve
   * @param obj the choice of objective function
   * @param moves the maximum number of moves to allow
   * @throws ilog.concert.IloException if the model cannot be built
   */
  public Model(final Problem problem, final Objective obj, final int moves)
         throws IloException {
    prob = problem;
    nNodes = problem.getnNodes();
    maxMoves = moves;
    objective = obj;
    // Initialize the model.
    model = new IloCplex();
    // Create the variables.
    location = new IloNumVar[maxMoves][nNodes];
    first = new IloNumVar[maxMoves][nNodes];
    done = new IloNumVar[maxMoves];
    for (int t = 0; t < maxMoves; t++) {
      done[t] = model.boolVar("finish_at_" + t);
      for (int n = 0; n < nNodes; n++) {
        location[t][n] = model.boolVar("visiting_" + n + "_at_" + t);
        first[t][n] = model.boolVar("first_to_" + n + "_at_" + t);
      }
    }
    // Constraint: the robot can only be at one node in each time period.
    for (int t = 0; t < maxMoves; t++) {
      model.addEq(model.sum(location[t]), 1.0, "one_stop_" + t);
    }
    // Constraint: the robot starts at node 0.
    location[0][0].setLB(1.0);
    // Constraint: the robot ends at node 0.
    location[maxMoves - 1][0].setLB(1.0);
    // Constraint: every node other than 0 must be visited at least once.
    for (int n = 1; n < nNodes; n++) {
      IloLinearNumExpr expr = model.linearNumExpr();
      for (int t = 1; t < maxMoves; t++) {
        expr.addTerm(1.0, location[t][n]);
      }
      model.addGe(expr, 1, "must_visit_" + n);
    }
    // Constraint: to be at any node other than the origin, the robot must have
    // come from a neighboring node.
    for (int t = 1; t < maxMoves; t++) {
      for (int n = 0; n < nNodes; n++) {
        IloLinearNumExpr expr = model.linearNumExpr();
        for (int m : problem.getNeighbors(n)) {
          expr.addTerm(1.0, location[t - 1][m]);
        }
        // For node 0 only, you can "move" to node 0 (stay put).
        if (n == 0) {
          expr.addTerm(1.0, location[t - 1][0]);
        }
        model.addLe(location[t][n], expr, "neighbor_time_" + t + "_node_" + n);
      }
    }
    // Constraint: every node other than the origin has a unique first visit.
    for (int n = 1; n < nNodes; n++) {
      IloLinearNumExpr expr = model.linearNumExpr();
      for (int t = 1; t < maxMoves; t++) {
        expr.addTerm(1.0, first[t][n]);
      }
      model.addEq(expr, 1.0, "node_" + n + "_first_visit");
    }
    // Constraint: the first visit to the origin is at time 0.
    model.addEq(first[0][0], 1.0, "origin_first_visit");
    // Constraint: the first visit to a node requires being at the node.
    for (int t = 0; t < maxMoves; t++) {
      for (int n = 0; n < nNodes; n++) {
        model.addLe(first[t][n], location[t][n], "first_" + t + "_" + n);
      }
    }
    // Constraint: the robot must eventually be officially done.
    model.addEq(model.sum(done), 1.0, "must_be_done");
    // Constraint: to be done, the robot must be at node 0.
    for (int t = 0; t < maxMoves; t++) {
      model.addLe(done[t], location[t][0], "home_at_" + t);
    }
    // Constraint: to be done, there can be no more visits outside node 0.
    for (int t = 0; t < maxMoves - 1; t++) {
      for (int tau = t + 1; tau < maxMoves; tau++) {
        for (int n = 1; n < nNodes; n++) {
          model.addLe(model.sum(done[t], location[tau][n]), 1.0,
                      "done_" + t + "_" + tau + "_" + t);
        }
      }
    }
    // Build the objective expressions.
    distance = model.linearNumExpr();
    delay = model.linearNumExpr();
    for (int t = 0; t < maxMoves; t++) {
      distance.addTerm(t, done[t]);
    }
    for (int n = 0; n < nNodes; n++) {
      double wt = problem.getPriority(n);
      for (int t = 0; t < maxMoves; t++) {
        delay.addTerm(wt * t, first[t][n]);
      }
    }
    // Add the objective functions to the model.
    IloCplexMultiCriterionExpr objs;
    switch (objective) {
      case DELAY_DISTANCE:
        objs = model.staticLex(new IloNumExpr[]{delay, distance},
                               "delay, then distance");
        modelObjective = model.minimize(objs);
        model.add(modelObjective);
        break;
      case DISTANCE_DELAY:
        objs = model.staticLex(new IloNumExpr[]{distance, delay},
                               "distance, then delay");
        modelObjective = model.minimize(objs);
        model.add(modelObjective);
        break;
      case DELAY:
        model.addMinimize(delay);
        break;
      case DISTANCE:
      default: // Should not happen: just minimize distance.
        model.addMinimize(distance);
    }
  }

  /**
   * Bounds the distance the robot may travel.
   * @param maxDistance the maximum allowed distance
   * @throws IloException if CPLEX hiccups
   */
  public void boundDistance(final double maxDistance) throws IloException {
    model.addLe(distance, maxDistance, "distance_bound");
  }

  /**
   * Bounds the overall delay encountered by "customers".
   * @param maxDelay the maximum allowed delay
   * @throws IloException if CPLEX hiccups
   */
  public void boundDelay(final double maxDelay) throws IloException {
    model.addLe(delay, maxDelay, "delay_bound");
  }

  /**
   * Solves the model.
   * @param timeLimit the run time limit in seconds
   * @return the final solver status
   * @throws IloException if CPLEX blows up
   */
  public IloCplex.Status solve(final double timeLimit) throws IloException {
    model.setParam(IloCplex.DoubleParam.TimeLimit, timeLimit);
    model.solve();
    return model.getStatus();
  }

  /**
   * Gets the values of the two criterion functions.
   * @return the vector [distance, delay] for the final solution
   * @throws IloException if no solution is found
   */
  public double[] getObjectiveValue() throws IloException {
    return new double[]{model.getValue(distance), model.getValue(delay)};
  }

  /**
   * Gets the robot's movements as a string.
   * @return a string describing the robot's movements
   * @throws IloException if there is no solution or it cannot be decoded
   */
  public String getSolution() throws IloException {
    StringBuilder sb = new StringBuilder();
    for (int t = 0; t < maxMoves; t++) {
      double[] x = model.getValues(location[t]);
      String where = "nowhere";
      boolean firstVisit = false;
      int node = 0;
      for (int n = 0; n < nNodes; n++) {
        if (x[n] > HALF) {
          node = n;
          where = prob.getCoordinates(n);
          firstVisit = model.getValue(first[t][n]) > HALF;
          break;
        }
      }
      sb.append("At t = ").append(t).append(", robot is at ").append(where);
      if (firstVisit) {
        double p = prob.getPriority(node);
        sb.append(" (first visit, with penalty ")
          .append(t * p).append(")");
      }
      sb.append("\n");
    }
    return sb.toString();
  }

  /**
   * Closes the model.
   */
  @Override
  public void close() {
    model.close();
  }

}
