package robot;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

/**
 * Problem holds an instance of the routing problem.
 *
 * @author Paul A. Rubin (rubin@msu.edu)
 */
public final class Problem {
  /*
    Nodes are arranged in a rectangular grid, numbered in raster scan order
    from lower left (0) to upper right.
  */
  private final int nRows;          // number of rows in the grid
  private final int nCols;          // number of columns in the grid
  private final int nNodes;         // number of nodes
  private final int[][] node;       // coordinates of each node
  private final HashMap<Integer, HashSet<Integer>> neighbors;
                                    // indices of neighboring nodes
  private final double[] priority;  // priority weight of each node

  /**
   * Constructor for the problem instance in the ORSE post.
   */
  @SuppressWarnings({ "checkstyle:magicnumber" })
  public Problem() {
    nRows = 5;
    nCols = 5;
    nNodes = nRows * nCols;
    node = new int[nNodes][2];
    neighbors = new HashMap<>();
    priority = new double[]{
      // Priorities from bottom to top.
      2, 2, 3, 4, 2,
      1, 1, 2, 2, 1,
      1, 3, 2, 3, 1,
      1, 3, 4, 4, 2,
      2, 4, 1, 2, 1
    };
    // Generate the nodes.
    int next = 0;
    for (int c = 0; c < nCols; c++) {
      for (int r = 0; r < nRows; r++) {
        node[next][0] = r;
        node[next][1] = c;
        next += 1;
      }
    }
    // Tally the neighbors.
    for (int n = 0; n < nNodes; n++) {
      neighbors.put(n, new HashSet<>());
    }
    for (int n = 0; n < nNodes; n++) {
      int x = node[n][0];
      int y = node[n][1];
      for (int m = n + 1; m < nNodes; m++) {
        if (Math.abs(node[m][0] - x) + Math.abs(node[m][1] - y) == 1) {
          neighbors.get(n).add(m);
          neighbors.get(m).add(n);
        }
      }
    }
  }

  /**
   * Gets an unmodifiable copy of the set of neighbors of a node.
   * @param n the index of the target node
   * @return the set indices of neighboring nodes
   */
  public Set<Integer> getNeighbors(final int n) {
    return Collections.unmodifiableSet(neighbors.get(n));
  }

  /**
   * Gets the priority of a node.
   * @param n the index of the target node
   * @return the priority of the node
   */
  public double getPriority(final int n) {
    return priority[n];
  }

  /**
   * Gets the number of nodes.
   * @return the number of nodes
   */
  public int getnNodes() {
    return nNodes;
  }

  /**
   * Gets the grid coordinates of a node (as a string).
   * @param n the index of the node
   * @return a string giving the node's coordinates
   */
  public String getCoordinates(final int n) {
    return "(" + node[n][0] + ", " + node[n][1] + ")";
  }
}
