package robot;

import ilog.concert.IloException;
import ilog.cplex.IloCplex;

/**
 * Robot tests out integer programming models for a prioritized routing problem
 * posed on OR Stack Exchange.
 *
 * https://or.stackexchange.com/questions/12451/coverage-path-planning-dilemma
 * -trade-off
 *
 * @author Paul A. Rubin (rubin@msu.edu)
 */
public final class Robot {

  /**
   * Dummy constructor.
   */
  private Robot() { }

  /**
   * Sets up and solves a problem instance.
   * @param args the command line arguments
   */
  @SuppressWarnings({ "checkstyle:magicnumber" })
  public static void main(final String[] args) {
    // Generate the test problem.
    Problem problem = new Problem();
    // Set some limits.
    int maxMoves = 75;
    double timeLimit = 300;
    double d1 = 0;
    double d2 = 0;
    // Build and solve a model prioritizing distance.
    try (Model model = new Model(problem, Objective.DISTANCE_DELAY, maxMoves)) {
      System.out.println("\n=== Solving a lexicographic model prioritizing"
                         + " distance. ===");
      IloCplex.Status status = model.solve(timeLimit);
      d1 = report(status, model);
    } catch (IloException ex) {
      System.out.println("CPLEX blew up!\n" + ex.getMessage());
    }
    // Now build and solve a model prioritizing delay.
    try (Model model = new Model(problem, Objective.DELAY_DISTANCE, maxMoves)) {
      System.out.println("\n=== Solving a lexicographic model prioritizing"
                         + " delay. ===");
      IloCplex.Status status = model.solve(timeLimit);
      d2 = report(status, model);
    } catch (IloException ex) {
      System.out.println("CPLEX blew up!\n" + ex.getMessage());
    }
    // Minimize delay without regard to distance. This should produce the same
    // delay as in the previous case.
    try (Model model = new Model(problem, Objective.DELAY, maxMoves)) {
      System.out.println("\n=== Solving a model minimizing delay. ===");
      IloCplex.Status status = model.solve(timeLimit);
      report(status, model);
    } catch (IloException ex) {
      System.out.println("CPLEX blew up!\n" + ex.getMessage());
    }
    // Minimize delay with a limit on distance that is between the minimum
    // possible distance and the distance from a lexicographic solution
    // prioritizing delay.
    try (Model model = new Model(problem, Objective.DELAY, maxMoves)) {
      System.out.println("\n=== Solving a model minimizing delay with"
                         + " limited distance. ===");
      model.boundDistance((d1 + d2) / 2);
      IloCplex.Status status = model.solve(timeLimit);
      report(status, model);
    } catch (IloException ex) {
      System.out.println("CPLEX blew up!\n" + ex.getMessage());
    }
  }

  /**
   * Reports the output of a model.
   * @param status the final solver status
   * @param model the model
   * @return the distance traveled in the solution
   * @throws IloException if the model cannot be interrogated for a solution
   */
  private static double report(final IloCplex.Status status, final Model model)
                        throws IloException {
      System.out.println("Final status = " + status + ".");
      double[] objValues = model.getObjectiveValue();
      System.out.println("Distance traveled = " + objValues[0] + ".");
      System.out.println("Delay score = " + objValues[1] + ".");
      System.out.println(model.getSolution());
      return(objValues[0]);
  }

}
